#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include "header.h"
#include "tools.h"
#include "util.h"
#include "globals.h"
#include "headerx.h"

/* Here is a list of supported ".var" file extension variations */
static const char *var_file_exts[] = {
  ".var",
  ".VAR"
};

int
LabelHasByteOffsetFields(LABEL *label)
{
  FIELD **fields;
  int     nfields;
  int     i;
  
  fields = (FIELD **)list_data(label->fields);
  nfields = list_count(label->fields);

  for (i = 0; i < nfields; i++){
    if (fields[i]->vardata != NULL){ /* We have a BYTE_OFFSET field */
		/* eformat == BYTE_OFFSET not available, as an eformat of
			BYTE_OFFSET is filtered by the label processing routines */
      return 1;
    }
  }

  return 0;
}

int
TableHasByteOffsetFields(TABLE *table)
{
  return LabelHasByteOffsetFields(table->label);
}


char *
GetVarFileName(const char *DatFileName)
{
  int i, ct;
  char *VarFileName;
  struct stat stat_buf;

  VarFileName = strdup(DatFileName);
  if (VarFileName == NULL){ return NULL; }

  ct = sizeof(var_file_exts) / sizeof(char *);
  for (i = 0; i < ct; i++){
    strcpy(strrchr(VarFileName, '.'), var_file_exts[i]);
    if ((stat(VarFileName, &stat_buf) > -1) &&
        (stat_buf.st_mode & S_IFREG)){
      return VarFileName;
    }
  }

  free(VarFileName);
  return NULL;
}


char *
MakeVarFileName(const char *DatFileName)
{
  char *VarFileName;

  VarFileName = strdup(DatFileName);
  if (VarFileName != NULL){
    if (!strrchr(VarFileName, '.')){
      free(VarFileName);
      return NULL;
    }
    strcpy(strrchr(VarFileName, '.'), var_file_exts[0]);
  }

  return VarFileName;
}

/**
 ** Function: PackVarRecs()
 **
 ** Weeds out variable records not being pointed to after the merge
 ** operation is complete. Note that variable records in the ".var"
 ** file are pointed to by the BYTE_OFFSET fields in ".dat" files.
 ** ".var" files cannot be used by themselves.
 **
 ** Notes:
 **   As a side-effect the fixed-records (as pointed to by
 **   tbl_buff->data) are updated to point to the new offset of 
 **   the variable data.
 **/
char *
PackVarData(
    char  *varfname,    /* Name of the ".var" file to be packed */
    char  *datfname,    /* Name of the associated ".dat" file - containing the fixed-size records */
    int    offset,      /* Offset value from the FRAGMENT header */
    int    nrows,       /* Number of rows as given in the FRAGMENT header */
    void  *frag_data,   /* ".dat" file loaded into memory */
    LABEL *label,       /* ".dat" file label */
    int    verbose      /* verbose mode */
)
{
  int     varfd;
  int     tmp_varfd;
  char   *tmp_varfname;
  int     i, j;
  FIELD **fields;
  int     nfields;
  FIELD  *field;
  char   *rec_data_ptr;
  char   *field_data_ptr;
  DATA    byte_offset;
  off_t   new_offset;
  char   *var_rec = NULL;
  size_t  var_rec_buff_sz = 0;
  int     nels;
  size_t  bytes_out;

  if (verbose){
    fprintf(stderr, "%s: Packing %s...\n", pgm_name, varfname);
  }

  if ((varfd = open(varfname, O_RDONLY)) < 0){
    fprintf(stderr, "%s: Cannot open file %s. ", pgm_name, varfname);
    perror("Reason");
    return NULL;
  }
  

  if ((tmp_varfd = CreateTmpFile(&tmp_varfname, basename(varfname))) < 0){
    fprintf(stderr, "%s: Cannot create temporary file for %s. ",
            pgm_name, varfname);
    perror("Reason");
    close(varfd);
    return NULL;
  }

  nfields = list_count(label->fields);
  fields = (FIELD **)list_data(label->fields);

  for (i = 0; i < nrows; i++){
    rec_data_ptr = GetRecPtr(frag_data, offset, label->reclen, i);
    
    /* Scan through all the table-fields or columns */
    for (j = 0; j < nfields; j++){
      field = fields[j];
      
      /* field->eformat == BYTE_OFFSET is not possible below since the
         label processing routine replaces the eformat from BYTE_OFFSET
         to MSB_INTEGER */
      if (field->vardata != NULL){
        /* If the field is a variable data field, then copy it to
           the temporary file and update the byte offset */
        field_data_ptr = GetFieldPtr(rec_data_ptr, field, 0);
        byte_offset = ConvertData(field_data_ptr, field);

        if (byte_offset.i ==  BYTE_OFFSET_NO_VAR_DATA){
          /* If byte-offset value is -1 (sentinal), then the field in 
             question does not have variable data associated with it
             for this row. So, replicate the sentinal to the output
             ".dat" record (residing in memory) */
          new_offset = BYTE_OFFSET_NO_VAR_DATA;
        }
        else {
          /* Otherwise, read the variable data record */
          nels = ReadVardataRec(varfd, byte_offset.i, field, &var_rec, &var_rec_buff_sz);
          if (nels < 0){
            fprintf(stderr,
                    "%s: Could not read vardata from file %s @ offset 0x%lX for field %s in rec#%d in %s.\n",
                    pgm_name, varfname, byte_offset.i, field->name, i, datfname);
            close(varfd);
            close(tmp_varfd); unlink(tmp_varfname); free(tmp_varfname);
            if (var_rec){ free(var_rec); }
            return NULL;
          }

          /* Get the new variable data byte-offset from the new/temp file */
          if ((new_offset = lseek(tmp_varfd, 0, SEEK_CUR)) < 0){
            fprintf(stderr, "%s: Seek failed in temporary file %s, while processing %s rec#%d. ",
                    pgm_name, varfname, datfname, i);
            perror("Reason");
            close(varfd);
            close(tmp_varfd); unlink(tmp_varfname); free(tmp_varfname);
            if (var_rec){ free(var_rec); }
            return NULL;
          }

          /* Write the variable data to the new/temp file. Note that we don't care for
				 the data format, since we are writing in the same format as we are reading
				 data in. */
          bytes_out = write(tmp_varfd, var_rec, VardataBuffSize(field->vardata, nels));
          /* bytes_out = write(tmp_varfd, var_rec, nels * field->vardata->size + 2 * sizeof(ushort)); */
          
          if (bytes_out != VardataBuffSize(field->vardata, nels)){
            fprintf(stderr, "%s: Write failed on temporary file %s, while processing %s. ",
                    pgm_name, varfname, datfname);
            perror("Reason");
            close(varfd);
            close(tmp_varfd); unlink(tmp_varfname); free(tmp_varfname);
            if (var_rec){ free(var_rec); }
            return NULL;
          }
        }

        byte_offset.i = new_offset;
        /* Stick in the new byte offset in the fixed part of the record */
        Int2ByteOffset(byte_offset, field_data_ptr, field->size, NULL, NULL);
      }
    }
  }

  if (var_rec){ free(var_rec); }
  close(varfd);
  close(tmp_varfd);

  if (verbose){
    fprintf(stderr, "%s: %s packing completed.\n", pgm_name, varfname);
  }

  /* Return the tmp vardata file name, which contain the packed vardata records.
     Note that the file name is allocated using malloc() */
  return tmp_varfname;
}
