/**
 ** Module: ReadVardataRec(fd, offset, *field, **data, *data_buff_sz)
 **
 ** Input:
 **
 **   fd - Open descriptor of the variable (or ".var") file
 **
 **   offset - A byte-offset value within the file, where the
 **       the variable record starts (i.e. its header)
 **
 **   field - Pointer to the byte-offset field associated with the
 **       variable data.
 **
 ** Input/Output:
 **
 **   data - Double pointer to the buffer where the data is returned.
 **       If *data == NULL, space will be allocated to it using
 **       malloc(). If *data != NULL, then the buffer may be extended
 **       using realloc() if so required. If *data == NULL, the
 **       data_buff_sz value is considered zero.
 **
 **   data_buff_sz - Pointer to the data buffer size. If zero, then
 **       *data is considered NULL. It contains the current 
 **       allocation of the *data buffer.
 **
 ** Return value:
 **   successful - returns the number of data elements in the
 **       data portion of the variable record.
 **   unsuccessful - returns -1
 **   
 ** This function reads in a record from the variable data or ".var"
 ** file (associated with a data or ".dat" file in a vanilla dataset).
 **
 ** The variable data record has the following format:
 **
 ** +---------------+------------------------------+---------------+
 ** | Size in bytes |  Data                        | Size in bytes |
 ** +---------------+------------------------------+---------------+
 **    (Header)                                        (Trailer)
 **
 ** The size in bytes is the size of the "data" portion only.
 ** Currently the size is limited to an unsigned short value.
 ** Whenever we talk about the variable data format, it is implictly
 ** assumed that the size in bytes value will always be a unsigned
 ** short BINARY value, nothing else.
 ** 
 ** A valid record will have the same byte count in the header as
 ** well as the trailer.
 **
 ** At the end of successful call (to this function), the file 
 ** pointer will be moved beyond the end of the variable data
 ** record.
 **/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include "globals.h"
#include "header.h"
#include "proto.h"
#include "headerx.h"

#define INITIAL_AUX_BUFF_SZ  8192

int
ReadVardataRec(int fd, off_t offset, FIELD *field, char **data, int *data_buff_sz)
{
  int     nels_hdr; /* no of vardata elements as read from header */
  int     nels_tlr; /* no of vardata elements as read from trailer */
  size_t  bytes_read; /* no of bytes read from the file */
  size_t  reqd_buff_sz; /* actual buffer space req'd to store this vardata record */
  char   *tlr_ptr; /* Pointer to vardata trailer */

  if (*data == NULL || *data_buff_sz == 0){
    /* If existing data buffer is empty, allocate new space */
    *data = (char *)calloc(1, INITIAL_AUX_BUFF_SZ);
    if (*data == NULL){
      fprintf(stderr,
              "%s: Cannot alloc %ld bytes mem for variable data at offset 0x%X.\n", 
              pgm_name, INITIAL_AUX_BUFF_SZ, offset);
      return -1;
    }
    *data_buff_sz = INITIAL_AUX_BUFF_SZ;
  }

  lseek(fd, offset, SEEK_SET); /* Move to start of variable data record */

  /* Read a buffer-size amount of data in the hope that the buffer
		will be big enough to hold all of the variable record */
  bytes_read = read(fd, *data, *data_buff_sz);
  
  /* Extract the number of elements from the variable data header
     We assume that this value is a short int */
  nels_hdr = (int)ConvertVarHdrBytes(*data) / field->vardata->size;
  /* assert(nels_hdr * field->vardata->size != (int)ConvertVarHdrBytes(*data)); */

  /* Buffer should be big enough to hold the variable data as well as the
     two shorts giving the variable data length (as heder and trailer) */
  reqd_buff_sz = VardataBuffSize(field->vardata, nels_hdr);
  /* reqd_buff_sz = nels_hdr * field->vardata->size + sizeof(ushort) * 2; */
  
  if (reqd_buff_sz > *data_buff_sz){    /* data outran the buffer */
    void *tmp = *data;

    /* If the data outruns the buffer size, then extend the buffer to
       accomodate the variable data record */
    *data = (char *)realloc(*data, reqd_buff_sz);
    if (*data == NULL){
      fprintf(stderr,
              "%s: Cannot alloc mem for variable data at offset 0x%X of size %ld.\n", 
              pgm_name, offset, reqd_buff_sz);
      *data = tmp;
      return -2; /* >>> Should I abort() here <<< */
    }

    *data_buff_sz = reqd_buff_sz;

    /* Read the rest of data for the current record */
    bytes_read += read(fd, (*data) + bytes_read, reqd_buff_sz - bytes_read);
  }

  if (bytes_read < reqd_buff_sz){
    /* If we couldn't read all of the record , declare an error */
    fprintf(stderr,
            "%s: Read of variable data record of size %ld bytes failed at offset 0x%X.\n",
            pgm_name, nels_hdr, offset);
    return -1; /* >>> Should I abort() here <<< */
  }
  
  /* Change to a more generic conversion */
  tlr_ptr = GetVFieldPtr(*data, field->vardata, nels_hdr);
  nels_tlr = (int)ConvertVarHdrBytes(tlr_ptr) / field->vardata->size;
  /* + sizeof(ushort) + nels_hdr * field->vardata->size) */
  if (nels_hdr != nels_tlr){
    /* Different header and trailer data sizes means corrupted variable data file. */
    fprintf(stderr,
            "%s: Variable data at offet 0x%X has different header and trailer counts, i.e. %ld and %ld respectively.\n",
            pgm_name, offset, nels_hdr, nels_tlr);
    /* >>> return -1;  What to do now - I think, let the user continue <<< */
  }

  return nels_hdr;
}



